/*
 * EffecTV - Realtime Digital Video Effector
 * Copyright (C) 2001 FUKUCHI Kentarou
 *
 * FireTV - clips incoming objects and burn them.
 * Copyright (C) 2001 FUKUCHI Kentarou
 *
 * Fire routine is taken from Frank Jan Sorensen's demo program.
 */

/* modified for Livido by G. Finch (salsaman)
   modifications (c) G. Finch */

#include "plugins-livido.h"
#include "livido-lib.c"
#include "math.h"


//////////////////////////////////

livido_setup_f livido_setup;

livido_init_f fire_init;
livido_deinit_f fire_deinit;
livido_process_f fire_process;

typedef unsigned int RGB32;

/////////////////////////////////

livido_palette_t in_palettes[]={LIVIDO_PALETTE_RGB32,LIVIDO_PALETTE_END};
livido_palette_t out_palettes[]={LIVIDO_PALETTE_RGB32,LIVIDO_PALETTE_END};

livido_channel_template_t in_channels={"In 1",0,in_palettes,0,0,0,NULL};
livido_channel_template_t out_channels={"Out 1",LIVIDO_CHANNEL_SAME_AS_SIZE|LIVIDO_CHANNEL_SAME_AS_PALETTE,out_palettes,1,0,0,NULL};

livido_instance_template_t fx={"fireTV","effectTV","Burn baby burn !",1,1.7,LIVIDO_PROPERTY_REALTIME,NULL,NULL,&in_channels,&out_channels,fire_init,fire_deinit,fire_process,NULL,NULL,NULL};


static RGB32 palette[256];

#define MaxColor 120
#define Decay 15
#define MAGIC_THRESHOLD 50


struct _sdata {
  unsigned char *buffer;
  short *background;
  unsigned char *diff;
  int threshold;
  uint32_t fastrand_val;
};


//////////////////////////////////////////////////////

inline uint32_t fastrand(struct _sdata *sdata)
{
#define rand_a 1073741789L
#define rand_c 32749L

  return (sdata->fastrand_val = rand_a * sdata->fastrand_val + rand_c);
}



static void HSItoRGB(double H, double S, double I, int *r, int *g, int *b)
{
  double T,Rv,Gv,Bv;
  
  T=H;
  Rv=1+S*sin(T-2*M_PI/3);
  Gv=1+S*sin(T);
  Bv=1+S*sin(T+2*M_PI/3);
  T=255.999*I/2;
  *r=trunc(Rv*T);
  *g=trunc(Gv*T);
  *b=trunc(Bv*T);
}


static void makePalette()
{
  int i, r, g, b;

  for(i=0; i<MaxColor; i++) {
    HSItoRGB(4.6-1.5*i/MaxColor, (double)i/MaxColor, (double)i/MaxColor, &r, &g, &b);
    palette[i] = (r<<16)|(g<<8)|b;
  }
  for(i=MaxColor; i<256; i++) {
    if(r<255)r++;if(r<255)r++;if(r<255)r++;
    if(g<255)g++;
    if(g<255)g++;
    if(b<255)b++;
    if(b<255)b++;
    palette[i] = (r<<16)|(g<<8)|b;
  }
}


static void image_bgsubtract_y(RGB32 *src, int width, int height, struct _sdata *sdata)
{
  register int i;
  int R, G, B;
  RGB32 *p;
  short *q;
  unsigned char *r;
  int v;
  int video_area=width*height;
  
  p = src;
  q = sdata->background;
  r = sdata->diff;
  for(i=0; i<video_area; i++) {
    R = ((*p)&0xff0000)>>(16-1);
    G = ((*p)&0xff00)>>(8-2);
    B = (*p)&0xff;
    v = (R + G + B) - (int)(*q);
    *r = ((v + sdata->threshold)>>24) | ((sdata->threshold - v)>>24);
    
    p++;
    q++;
    r++;
  }
  
  /* The origin of subtraction function is;
   * diff(src, dest) = (abs(src - dest) > threshold) ? 0xff : 0;
   *
   * This functions is transformed to;
   * (threshold > (src - dest) > -threshold) ? 0 : 0xff;
   *
   * (v + threshold)>>24 is 0xff when v is less than -threshold.
   * (v - threshold)>>24 is 0xff when v is less than threshold.
   * So, ((v + threshold)>>24) | ((threshold - v)>>24) will become 0xff when
   * abs(src - dest) > threshold.
   */
}




livido_instance_template_t *livido_setup(void) {
  makePalette();
  return &fx;
}



int fire_init(livido_instance_t *inst) {

  int map_h;
  int map_w;
  struct _sdata *sdata;

  if (inst==NULL) return LIVIDO_ERROR_NULL_INSTANCE;

  map_h=inst->in_channels[0].height;
  map_w=inst->in_channels[0].width;

  sdata=malloc(sizeof(struct _sdata));

  if(sdata == NULL ) {
    return LIVIDO_ERROR_MEMORY_ALLOCATION;
  }

  sdata->buffer = (unsigned char *)malloc(map_h*map_w*sizeof(unsigned char));
  if(sdata->buffer == NULL ) {
    free(sdata);
    return LIVIDO_ERROR_MEMORY_ALLOCATION;
  }
  sdata->background = (short *)malloc(map_h*map_w*sizeof(short));
  if(sdata->background == NULL ) {
    free(sdata->buffer);
    free(sdata);
    return LIVIDO_ERROR_MEMORY_ALLOCATION;
  }
  sdata->diff = (unsigned char *)malloc(map_h*map_w * sizeof(unsigned char));
  if(sdata->diff == NULL ) {
    free(sdata->background);
    free(sdata->buffer);
    free(sdata);
    return LIVIDO_ERROR_MEMORY_ALLOCATION;
  }
  sdata->threshold=MAGIC_THRESHOLD*7;
  memset(sdata->buffer, 0, map_h*map_w*sizeof(unsigned char));
  sdata->fastrand_val=0;

  inst->internal=sdata;
  
  return LIVIDO_NO_ERROR;
}


int fire_deinit(livido_instance_t *inst) {
  struct _sdata *sdata;
  if (inst==NULL) return LIVIDO_ERROR_NULL_INSTANCE;
  sdata=inst->internal;
  if (sdata != NULL) {
    free(sdata->buffer);
    free(sdata->diff);
    free(sdata->background);
    free(sdata);
  }
  return LIVIDO_NO_ERROR;
}






int fire_process (livido_instance_t *inst, livido_frame_t *in, livido_frame_t *out) {
  RGB32 *src,*dest;
  struct _sdata *sdata;
  int video_width,video_height;
  register int i, x, y;
  unsigned char v;
  int video_area;

  if (inst==NULL) return LIVIDO_ERROR_NULL_INSTANCE;

  sdata=inst->internal;
  src=in->pixel_data[0];
  dest=out->pixel_data[0];

  video_width = inst->in_channels[0].width;
  video_height = inst->in_channels[0].height;
  video_area=video_width*video_height;

  image_bgsubtract_y(src,video_width,video_height,sdata);
  for(i=0; i<video_area-video_width; i++) {
    sdata->buffer[i] |= sdata->diff[i];
  }
  for(x=1; x<video_width-1; x++) {
    i = video_width + x;
    for(y=1; y<video_height; y++) {
      v = sdata->buffer[i];
      if(v<Decay)
	sdata->buffer[i-video_width] = 0;
      else
      sdata->buffer[i-video_width+fastrand(sdata)%3-1] = v - (fastrand(sdata)&Decay);
      i += video_width;
    }
  }
  for(y=0; y<video_area; y+=video_width) {
    for(x=1; x<video_width-1; x++) {
      dest[y+x] = palette[sdata->buffer[y+x]];
    }
  }
  return LIVIDO_NO_ERROR;
}
